@@ -4,6 +4,7 @@ import logging
 import re
 import sys
 from urllib.parse import unquote
+from xml.dom import minidom
 
 import requests
 from bs4 import BeautifulSoup
@@ -12,6 +13,9 @@ from requests.auth import HTTPBasicAuth
 from easymigration.config import init
 
 
+NS_EAS = "http://easy.dans.knaw.nl/easy/easymetadata/eas/"
+
+
 def update_thematische_collecties(fedora_config):
     """
      Copies a Thematische-Collecties.csv from stdin to stdout.
@@ -36,24 +40,52 @@ def update_thematische_collecties(fedora_config):
             for dataset_id in row["EASY-dataset-id"].split(","):
                 try:
                     jumpoff_id = get_jumpoff_id(dataset_id, risearch_url, auth)
-                    row["members"] += get_members(jumpoff_id, base_url)
+                    if jumpoff_id == "":
+                        row["members"] += members_from_relations(dataset_id, base_url)
+                    else:
+                        row["members"] += members_from_jumpoff(jumpoff_id, base_url)
                 except Exception as e:
                     logging.error(f"{e} PROCESSING {row}")
         csv_writer.writerow(row)
 
 
-def get_members(jumpoff_id, base_url):
-    mu = get_jumpoff_content(jumpoff_id, "HTML_MU", base_url)
-    logging.debug(f"{jumpoff_id}")
-    if mu.status_code == 404:
-        logging.debug(f"{jumpoff_id}")
-        mu = get_jumpoff_content(jumpoff_id, "TXT_MU", base_url)
-    if 200 == mu.status_code:
-        dataset_ids = parse_jumpoff(mu.text, jumpoff_id)
-        return ",".join(dataset_ids).replace("[]", "\"")
+def members_from_relations(dataset_id, base_url):
+    logging.debug(f"no jumpoff, trying EMD of {dataset_id}")
+    emd = get_data_stream(dataset_id, "EMD", base_url)
+    members = []
+    if emd.status_code != 200:
+        logging.error(f"could not read EMD {dataset_id} {emd.status_code}")
     else:
-        logging.error(f"jumpoff not found {jumpoff_id} {mu.status_code}")
-        return ""
+        relations = minidom.parseString(emd.content).getElementsByTagNameNS(NS_EAS, "relation")
+        members = resolve(map(extract_subject_link, filter(has_emphasis, relations)), dataset_id)
+    return ",".join(members)
+
+
+def extract_subject_link(relation):
+    return relation.getElementsByTagNameNS(NS_EAS, "subject-link")[0].firstChild.nodeValue
+
+
+def has_emphasis(relation):
+    return relation.getAttributeNS(NS_EAS, "emphasis") == "true"
+
+
+def members_from_jumpoff(jumpoff_id, base_url):
+    logging.debug(f"reading HTML_MU {jumpoff_id}")
+    mu = get_data_stream(jumpoff_id, "HTML_MU", base_url)
+    if mu.status_code == 404:
+        logging.debug(f"reading TXT_MU {jumpoff_id}")
+        mu = get_data_stream(jumpoff_id, "TXT_MU", base_url)
+    dataset_ids = []
+    if 200 != mu.status_code:
+        logging.error(f"could not read jumpoff {jumpoff_id} {mu.status_code}")
+    else:
+        soup = BeautifulSoup(mu.text, "html.parser")
+        dataset_ids = resolve(map(extract_href, soup.findAll("a")), jumpoff_id)
+    return ",".join(dataset_ids).replace("[]", "\"")
+
+
+def extract_href(node):
+    return unquote(node.attrs.get("href"))
 
 
 def get_jumpoff_id(dataset_id, risearch_url, auth):
@@ -66,43 +98,44 @@ def get_jumpoff_id(dataset_id, risearch_url, auth):
     params["format"] = "CSV"
 
     response = requests.get(risearch_url, params=params, auth=auth)
-    if response.status_code != 200 or "dans-jumpoff" not in response.text:
-        raise Exception(f"Could not find jumpoff-id for {dataset_id} response: {response}")
-    return re.findall(r'dans-jumpoff:[0-9]+', response.text)[0]
+    if response.status_code != 200:
+        raise Exception(f"Could not find jumpoff-id for {dataset_id} response: {response} {response.content}")
+    elif "dans-jumpoff" not in response.text:
+        return ""
+    else:
+        return re.findall(r'dans-jumpoff:[0-9]+', response.text)[0]
 
 
-def get_jumpoff_content(jumpoff_id, response_format, base_url):
-    url = f"{base_url}/objects/{jumpoff_id}/datastreams/{response_format}/content"
+def get_data_stream(object_id, stream, base_url):
+    url = f"{base_url}/objects/{object_id}/datastreams/{stream}/content"
     response = requests.get(url)
     logging.debug(f"status code: {response.status_code} url: {url}")
     return response
 
 
-def parse_jumpoff(jumpoff_page, jumpoff_id):
-    # members: "easy-dataset:34099, easy-dataset:57698"
-    soup = BeautifulSoup(jumpoff_page, "html.parser")
+def resolve(urls, object_id):
     dataset_ids = set()
-    for a_tag in soup.findAll("a"):
-        href = unquote(a_tag.attrs.get("href"))
-        logging.debug(f"resolving {href} in {jumpoff_id}")
+    for href in urls:
+        logging.debug(f"resolving {href} in {object_id}")
         if "easy-dataset:" in href:
             dataset_ids.add(re.findall(r'easy-dataset:[0-9]+', href)[0])
             continue
         if re.search("(?s).*(doi.org.*dans|urn:nbn:nl:ui:13-).*", href) is None:
-            logging.debug(f"Not a dataset link {href} in {jumpoff_id}")
+            logging.debug(f"Not a dataset link {href} in {object_id}")
             continue
         replaced_href = re.sub("http://dx.doi.org", "https://doi.org", href)
         try:
             response = requests.get(replaced_href, allow_redirects=False, timeout=0.5)
         except Exception as e:
-            logging.error(f"Could not resolve {replaced_href} in {jumpoff_id}: {e}")
+            logging.error(f"Could not resolve {replaced_href} in {object_id}: {e}")
             continue
         if response.status_code != 302:
-            logging.error(f"Expected status code 302 but got {response.status_code} for {replaced_href} in {jumpoff_id}")
+            logging.error(
+                f"Expected status code 302 but got {response.status_code} for {replaced_href} in {object_id}")
             continue
         location = unquote(response.headers.get("location"))
         if "easy-dataset:" not in location:
-            logging.error(f"Need 'easy-dataset:NNN' but {replaced_href} in {jumpoff_id} resolved to {location}")
+            logging.error(f"Need 'easy-dataset:NNN' but {replaced_href} in {object_id} resolved to {location}")
             continue
         dataset_ids.add(re.findall(r'easy-dataset:[0-9]+', location)[0])
     return dataset_ids
@@ -115,6 +148,7 @@ def main():
         description='Copies an easy-convert-bag-to-deposit/src/main/assembly/dist/cfg/ThemathischeCollecties.csv '
                     'from stdin to stdout. '
                     'Empty member fields will be updated by collecting links from the jumpoff page of the dataset. '
+                    'If the dataset has no jumpoff, relations with emphasis are collected from the EMD. '
                     'The first line of the CSV is assumed to be a header and is copied as-is.',
         epilog='Example: update_thematische_collecties.py < OldThemCol.csv > NewThemCol.csv'
     ).parse_args()
